@@ -2,6 +2,7 @@ from __future__ import unicode_literals, print_function
 import os
 import sys
 import six
+import zlib
 import time
 import types
 import base64
@@ -14,16 +15,18 @@ from subprocess import Popen
 from subprocess import PIPE
 from subprocess import STDOUT
 from copy import deepcopy
+from ConfigObject import ConfigObject
 from contextlib import contextmanager
 
 __all__ = [
-    'console_script', 'sh', 'env', 'stdin', 'test',
+    'console_script', 'requires', 'sh', 'env', 'ini', 'stdin', 'test',
     'ls', 'cat', 'grep', 'find', 'cut', 'tr', 'head', 'tail', 'sed', 'awk',
     'nc', 'ping', 'nmap', 'hostname', 'host', 'scp', 'rsync', 'wget', 'curl',
     'cd', 'which', 'mktemp', 'echo', 'wc',
     'tar', 'gzip', 'gunzip', 'zip', 'unzip',
     'vlc', 'ffmpeg', 'convert',
     'virtualenv', 'pip',
+    'git', 'hg', 'svn',
     'ssh', 'sudo',
     'path', 'pwd',  # path is posixpath, pwd return os.getcwd()
     'escape', 'e',  # e is escape()
@@ -67,6 +70,14 @@ def escape(value):
     return value
 
 
+def ini(filename, **defaults):
+    """Load a .ini file in a ConfigObject. Dont raise if the file does not
+    exist"""
+    filename = os.path.expanduser(filename)
+    defaults.update(home=os.path.expanduser('~'))
+    return ConfigObject(filename=filename, defaults=defaults)
+
+
 class Environ(dict):
     """Manage os.environ"""
 
@@ -81,6 +92,11 @@ class Environ(dict):
             value = os.pathsep.join(value)
         self[attr.upper()] = value
 
+    def __delattr__(self, attr):
+        attr = attr.upper()
+        if attr in self:
+            del self[attr]
+
     def __call__(self, **kwargs):
         environ = self.__class__(self.copy())
         for k, v in kwargs.items():
@@ -320,7 +336,7 @@ class Pipe(object):
 
     def __iter__(self):
         for line in self.stdout:
-            yield self._decode(line)
+            yield self._decode(line).rstrip('\n')
 
     def __call__(self, **kwargs):
         if self._done and self._stdout is not None:
@@ -485,7 +501,7 @@ class Stdout(str):
 
     @property
     def stdout(self):
-        return str(self)
+        return self
 
 
 class PyPipe(Pipe):
@@ -663,108 +679,221 @@ def wraps_module(mod):
 #####################
 
 
-def console_script(*args, **docopts):
-    def _console_script(func, **docopts):
-        @functools.wraps(func)
-        def wrapper(arguments=None):
-            if 'help' not in docopts:
-                docopts['help'] = True
-            if 'doc' not in docopts:
-                doc = getattr(func, '__doc__', None)
+@contextmanager
+def requires(*requirements, **kwargs):
+    """Add extra dependencies in a virtualenv"""
+    if '/.tox/' in sys.executable:
+        venv = os.path.dirname(os.path.dirname(sys.executable))
+    elif env.virtual_env:
+        venv = env.chut_virtualenv = env.virtual_env
+    else:
+        venv = os.path.expanduser(kwargs.get('venv', '~/.chut/venv'))
+    if not env.pip_download_cache:
+        env.pip_download_cache = os.path.expanduser('~/.chut/cache')
+        sh.mkdir('-p', env.pip_download_cache)
+    bin_dir = os.path.join(venv, 'bin')
+    if bin_dir not in env.path:
+        env.path = [bin_dir] + env.path
+    requirements = list(requirements)
+    if 'chut' not in requirements:
+        requirements.insert(0, 'chut')
+    if not test.d(venv):
+        import urllib
+        url = 'https://raw.github.com/pypa/virtualenv/master/virtualenv.py'
+        urllib.urlretrieve(url, '/tmp/virtualenv.py')
+        sh[sys.executable]('-S /tmp/_virtualenv.py', venv) > 1
+        sh.rm('/tmp/_virtualenv*', shell=True)
+        print('Installing %s...' % ', '.join(requirements))
+        sh.pip('install -qM', *requirements) > 1
+    elif env.chut_virtualenv:
+        upgrade = '--upgrade-deps' in sys.argv
+        if (env.chut_upgrade or upgrade):
+            installed = ''
+        else:
+            installed = str(sh.pip('freeze')).lower()
+        requirements = [r for r in requirements if r.lower() not in installed]
+        if requirements:
+            print('Updating %s...' % ', '.join(requirements))
+            sh.pip('install -qM --upgrade', *requirements) > 1
+    executable = os.path.join(bin_dir, 'python')
+    if not env.chut_virtualenv:
+        env.chut_virtualenv = venv
+        env.CHUTIFIED_FILES = os.environ.get('CHUTIFIED_FILES', '')
+        os.execve(executable, [executable] + sys.argv, env)
+    yield True
+
+
+class console_script(object):
+    """A decorator to take care of sys.argv via docopt"""
+
+    def __init__(self, *args, **opts):
+        self._console_script = True
+        self.docopts = opts
+        self.func = self.doc = None
+        self.wraps(args)
+
+    def version(self):
+        version = getattr(sys.modules['__main__'], 'version', 'unknown')
+        print('%s %s' % (self.func.__name__, version))
+
+    def wraps(self, args):
+        if args:
+            self.func = args[0]
+            functools.wraps(self.func)(self)
+            if 'help' not in self.docopts:
+                self.docopts['help'] = True
+            if 'doc' not in self.docopts:
+                doc = getattr(self.func, '__doc__', None)
             else:
-                doc = docopts.pop('doc')
+                doc = self.docopts.pop('doc')
             if doc is None:
                 doc = 'Usage: %prog'
-            name = func.__name__.replace('_', '-')
+            name = self.func.__name__.replace('_', '-')
             doc = doc.replace('%prog', name).strip()
             doc = doc.replace('\n    ', '\n')
-            import docopt
-            if isinstance(arguments, list):
-                docopts['argv'] = arguments
-                arguments = docopt.docopt(doc, **docopts)
-                return func(arguments)
-            else:
-                arguments = docopt.docopt(doc, **docopts)
-                sys.exit(func(arguments))
-        wrapper.console_script = True
-        return wrapper
-    if not args:
-        def waiting_for_func(func):
-            return _console_script(func, **docopts)
-        return waiting_for_func
-    else:
-        return _console_script(args[0])
+            self.doc = doc
+        return self
+
+    def main(self, arguments=None):
+        import docopt
+        ret = isinstance(arguments, list)
+        if ret:
+            self.docopts['argv'] = arguments
+        arguments = docopt.docopt(self.doc, **self.docopts)
+        if arguments.get('--version') is True:
+            res = self.version()
+        else:
+            try:
+                res = self.func(arguments)
+            except KeyboardInterrupt:
+                sys.exit(1)
+        return res if ret else sys.exit(res)
+
+    def __call__(self, *args, **kwargs):
+        return self.main(*args, **kwargs) if self.func else self.wraps(args)
 
 
-def generate(filename, arguments=None):
-    if arguments is None:
-        arguments = {}
-    if not os.path.isfile(filename):
-        mod = __import__(filename, globals(), locals(), [''])
-        filename = mod.__file__
-        name = mod.__name__
-    else:
+class Generator(object):
+    """generate a script from a @console_script. args may contain some
+    docopts like arguments"""
+
+    _modules = {}
+    _env_key = str('CHUTIFIED_FILES')
+
+    def __init__(self, **args):
+        self.devel = args.get('--devel') or args.get('devel')
+        if self.devel:
+            dest = 'bin'
+        else:
+            dest = args.get('--destination') or args.get('destination')
+            dest = os.path.expanduser(dest or 'dist/scripts')
+        sh.mkdir('-p', dest)
+        self.dest = dest
+        args.update(version=repr(str(args.get('--version') or 'unknown')),
+                    interpreter=args.get('--interpreter', 'python'))
+        self.args = args
+        self.mods = self.encode_modules(*args.get('modules', []))
+
+    def encode_module(self, mod):
+        if not hasattr(mod, '__file__'):
+            mod = __import__(mod)
+        name = str(mod.__name__)
+        if name not in self._modules:
+            data = inspect.getsource(mod)
+            data = base64.encodestring(zlib.compress(six.b(data)))
+            code = '_chut_modules.append((%r, %r))\n' % (name, data)
+            self._modules[name] = code
+        return self._modules[name]
+
+    def encode_modules(self, *modules):
+        try:
+            # check if the script is already chutified
+            _chut_modules = sys.modules['__main__']._chut_modules
+        except AttributeError:
+            # get source from files
+            modules = [
+                'six', 'docopt', 'ConfigObject', sys.modules[__name__]
+            ] + list(modules)
+            modules = ''.join([self.encode_module(m) for m in modules])
+        else:
+            # get source from _chut_modules
+            modules = ''
+            for name, data in _chut_modules:
+                modules += '_chut_modules.append((%r, %r))\n' % (name, data)
+        return modules
+
+    def generate(self, filename, args=None, **kwargs):
+        chutified = os.environ.get(self._env_key, str(''))
+        if filename in chutified:
+            return []
+        if args is None:
+            args = {}
+        args.update(kwargs)
         dirname = os.path.dirname(filename)
         sys.path.insert(0, dirname)
         name = inspect.getmodulename(filename)
-        mod = __import__(name)
+        os.environ.update(env)
+        mod = __import__(name, {'os': os}, {'os': os})
+        os.environ[self._env_key] = str(chutified + filename + ':')
 
-    filenames = []
-    for k, v in mod.__dict__.items():
-        if getattr(v, 'console_script', False) is True:
-            filenames.append(k)
+        console_scripts = []
+        for k, v in mod.__dict__.items():
+            if getattr(v, '_console_script', False) is True:
+                console_scripts.append(k)
 
-    if arguments.get('--list-entry-points'):
-        for name in filenames:
-            print(('%s = %s:%s' % (name.replace('_', '-'),
-                                   mod.__name__, name)))
-        return 0
+        scripts = []
+        for name in console_scripts:
+            script = os.path.join(self.dest, name.replace('_', '-'))
+            with open(script, 'w') as fd:
+                fd.write(SCRIPT_HEADER % self.args + self.mods + LOAD_MODULES)
+                if self.devel:
+                    fd.write('sys.path.insert(0, "%s")\n' % dirname)
+                    fd.write('import %s\n' % mod.__name__)
+                    fd.write('if __name__ == "__main__":\n')
+                    fd.write('    %s.%s()\n' % (mod.__name__, name))
+                else:
+                    fd.write(inspect.getsource(mod).replace('__main__',
+                                                            '__chutified__'))
+                    fd.write("\nif __name__ == '__main__':\n    %s()\n" % name)
+            executable = sh.chmod('+x', script)
+            if executable:
+                print(executable.commands_line)
+                scripts.append(script)
+            else:
+                print('failed to generate %s' % script)
+        return scripts
 
-    dest = os.path.expanduser(arguments.get('--destination', 'dist/scripts'))
-    sh.mkdir('-p', dest)
-
-    def encode_module(mod):
-        if not hasattr(mod, '__file__'):
-            mod = __import__(mod)
-        data = inspect.getsource(mod)
-        data = base64.encodestring(six.b(data))
-        return 'mods.append((%r, %r))\n' % (str(mod.__name__), data)
-
-    modules = [
-        'six', 'docopt', 'ConfigObject', sys.modules[__name__]
-    ] + arguments.get('<MODULE>', [])
-    modules = ''.join([encode_module(m) for m in modules])
-
-    for name in filenames:
-        script = os.path.join(dest, name.replace('_', '-'))
-        with open(script, 'w') as fd:
-            fd.write(SCRIPT_HEADER + modules + LOAD_MODULES)
-            fd.write(inspect.getsource(mod).replace('__main__',
-                                                    '__chutified__'))
-            fd.write("if __name__ == '__main__':\n    %s()\n" % name)
-        executable = sh.chmod('+x', script)
-        if executable:
-            print(executable.commands_line)
-        else:
-            print('failed to generate %s' % script)
-    return 0
+    def __call__(self, location):
+        scripts = []
+        if os.path.isfile(location):
+            scripts.extend(self.generate(location))
+        elif os.path.isdir(location):
+            filenames = sh.grep('-lRE --include=*.py @.*console_script',
+                                location) | sh.grep('-v site-packages')
+            for filename in sorted(filenames):
+                scripts.extend(self.generate(filename))
+        return scripts
 
 
 SCRIPT_HEADER = '''
-#!/usr/bin/env python
-import base64, json, types, sys
+#!/usr/bin/env %(interpreter)s
+
+version = %(version)s
+
+import base64, json, types, zlib, sys, os
+os.environ['CHUTIFIED'] = '1'
 PY3 = sys.version_info[0] == 3
-mods = []
+_chut_modules = []
 '''.lstrip()
 
 LOAD_MODULES = '''
-for name, code in mods:
+for name, code in _chut_modules:
     if PY3:
         if isinstance(code, str):
             code = code.encode('utf-8')
     else:
         name = bytes(name)
-    code = base64.decodestring(code)
+    code = zlib.decompress(base64.decodestring(code))
     mod = types.ModuleType(name)
     globs = dict()
     if PY3:
@@ -778,8 +907,7 @@ for name, code in mods:
         mod.wraps_module(mod)
     else:
         sys.modules[name] = mod
-
-import six
+from chut import env
 '''.lstrip()
 
 if __name__ != '__main__':
